# Fortuna: Random Value Generator
Fortuna's main goal is to provide a quick and easy way to build custom random-value generators for your data. While performance is not the primary focus, Fortuna is about an order of magnitude faster than the builtin Random module.

The core functionality of Fortuna is based on the RNG Storm engine. While Storm is a high quality random engine, Fortuna is not appropriate for cryptography of any kind. Fortuna is meant for games, data science, A.I. and experimental programming, not security.

Suggested Installation: `$ pip install Fortuna`

Installation on platforms other than MacOS may require building from source files.

Support this project: https://www.patreon.com/brokencode


### Documentation Table of Contents:
- Numeric Limits
- Project Terminology
- Random Generators:
    - Value Generators
        - `RandomValue(Collection) -> Callable -> Value`
        - `TruffleShuffle(Collection) -> Callable -> Value`
        - `QuantumMonty(Collection) -> Method Pack -> Value`
        - `CumulativeWeightedChoice(Table) -> Callable -> Value`
        - `RelativeWeightedChoice(Table) -> Callable -> Value`
        - `FlexCat(Matrix) -> Callable -> Value`
    - Integer Generators
        - `random_below(Integer) -> Integer`
        - `random_int(Integer, Integer) -> Integer`
        - `random_range(Integer, Integer, Integer) -> Integer`
        - `d(Integer) -> Integer`
        - `dice(Integer, Integer) -> Integer`
        - `plus_or_minus(Integer) -> Integer`
        - `plus_or_minus_linear(Integer) -> Integer`
        - `plus_or_minus_gauss(Integer) -> Integer`
    - Index Generators: ZeroCool Methods
        - ZeroCool Method Specification: `f(N) -> [0, N)` or `f(-N) -> [-N, 0)`
        - `random_index(Integer) -> Integer`
        - `front_gauss(Integer) -> Integer`
        - `middle_gauss(Integer) -> Integer`
        - `back_gauss(Integer) -> Integer`
        - `quantum_gauss(Integer) -> Integer`
        - `front_poisson(Integer) -> Integer`
        - `middle_poisson(Integer) -> Integer`
        - `back_poisson(Integer) -> Integer`
        - `quantum_poisson(Integer) -> Integer`
        - `front_geometric(Integer) -> Integer`
        - `middle_geometric(Integer) -> Integer`
        - `back_geometric(Integer) -> Integer`
        - `quantum_geometric(Integer) -> Integer`
        - `quantum_monty(Integer) -> Integer`
    - Float Generators
        - `canonical() -> Float`
        - `random_float(Float, Float) -> Float`
    - Bool Generator
        - `percent_true(Float) -> Boolean`
    - Shuffle Algorithms
        - `shuffle(List) -> None`
        - `knuth(List) -> None`
        - `fisher_yates(List) -> None`
- Test Suite
    - `distribution_timer(Callable, *args, **kwargs) -> None`
    - `quick_test() -> None`
- Development Log
- Test Suite Output
- Legal Information


#### Numeric Limits:
- Integer: 64 bit signed integer.
    - Input & Output Range: `(-2**63, 2**63)` or approximately +/- 9.2 billion billion.
    - Minimum: -9223372036854775807
    - Maximum:  9223372036854775807
- Float: 64 bit floating point.
    - Minimum: -1.7976931348623157e+308
    - Maximum:  1.7976931348623157e+308
    - Epsilon Below Zero: -5e-324
    - Epsilon Above Zero:  5e-324


#### Project Terminology:
- Value: Any python object.
- Callable: Any callable object, function, method or lambda.
- Collection: A group of Values.
    - List, Tuple, Set, etc... Any object that can be converted into a list via `list(some_object)`.
    - Comprehensions that produce a Collection also qualify.
    - Fortuna classes that wrap a Collection can wrap a Collection, Sequence or generator.
    - Fortuna functions that take a Collection as input will always require a Sequence.
- Sequence: An ordered Collection.
    - List, tuple or list comprehension.
    - A Sequence is an ordered Collection that can be indexed like a list, without conversion.
    - All Sequences are Collections but not all Collections are Sequences.
- Pair: Collection of two Values.
- Table: Collection of Pairs.
- Matrix: Dictionary of Collections.
- Inclusive Range.
    - `[1, 10] -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
- Exclusive Range.
    - `(0, 11) -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
- Partially Exclusive Range.
    - `[1, 11) -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
    - `(0, 10] -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
- Automatic Flattening.
    - All Random Value Generators in Fortuna will recursively call or "flatten" callable objects returned from the data at call time.
    - Static objects and non-default callable objects will be returned in an uncalled state without error.
    - A callable that can be flattened is any class initializer, function, method or lambda so long as it requires no arguments, it will be automatically flattened.
    - Mixing callable objects with un-callable objects is fully supported, but it can look a bit messy at first.
    - Nested callable objects are fully supported. Because `lambda(lambda) -> lambda` fixes everything for arbitrary values of 'because', 'fixes' and 'everything'.
    - To disable automatic flattening, pass the optional keyword argument `flat=False`.


## Random Value Generators

### Fortuna.random_value
`Fortuna.random_value(data: Sequence, flat=True) -> Value`
Drop in replacement for Random.choice()
- @param data :: Sequence of Values.
- @param flat :: Bool. Default: True. Option to flatten callable values.
- @return Value :: Produces a random value from the list with a flat uniform distribution.

```python
from Fortuna import random_value


apples = ["Delicious", "Honey Crisp", "Macintosh"]
fruit = [
    lambda: f"Apple, {random_value(apples)}",
    "Banana",
    "Cherry",
    "Grapes",
    "Orange",
]
apples.append("Granny Smith")  # Dynamic Modification
print(random_value(fruit))  # prints a random fruit, see the distribution below.

"""
>>> distribution_timer(random_value, fruit, label="random_value(fruit)")
Output Analysis: random_value(fruit)
Typical Timing: 219 ± 16 ns
Distribution of 10000 Samples:
 Apple, Delicious: 5.13%
 Apple, Granny Smith: 4.93%
 Apple, Honey Crisp: 4.9%
 Apple, Macintosh: 4.86%
 Banana: 20.16%
 Cherry: 20.01%
 Grapes: 19.63%
 Orange: 20.38%
"""
```


### TruffleShuffle
`Fortuna.TruffleShuffle(data: Collection, flat=True) -> Callable -> Value`
- Non-destructive, copies the data once.
- @param data :: Collection of Values. Set recommended.
- @param flat :: Bool. Default: True. Option to flatten callable values.
- @return :: Callable Instance.
    - @return :: Random value from the list with a **Wide Uniform Distribution**. The average width of the output distribution will naturally scale up with the size of the set.

**Wide Uniform Distribution**: *"Wide"* refers to the average distance between consecutive occurrences of the same value in the output sequence. The goal of this type of distribution is to keep the output sequence free of clumps or streaks of the same value, while maintaining randomness and uniform probability. This is not the same as a *flat uniform distribution*. The two distributions over time will be statistically similar for any given set, but the repetitiveness of the output sequence will be very different.

#### TruffleShuffle, Basic Use
```python
from Fortuna import TruffleShuffle


list_of_values = { 1, 2, 3, 4, 5, 6 }
truffle_shuffle = TruffleShuffle(list_of_values)

print(truffle_shuffle())  # will print one of the numbers 1-6

for _ in range(100):
    assert truffle_shuffle() is not truffle_shuffle()  # will never fire, except for a set of one.
```

#### TruffleShuffle with Flattening
```python
from Fortuna import TruffleShuffle


auto_flat = TruffleShuffle([lambda: 1, lambda: 2, lambda: 3])
print(auto_flat())  # will print the value 1, 2 or 3.
# Note: the lambda will not be called until call time and stays dynamic for the life of the object.

un_flat = TruffleShuffle([lambda: 1, lambda: 2, lambda: 3], flat=False)
print(un_flat()())  # will print the value 1, 2 or 3, mind the double-double parenthesis

auto_un_flat = TruffleShuffle([lambda x: x, lambda x: x + 1, lambda x:  x + 2], flat=False)
# Note: flat=False is not required here because the lambdas can not be called without input x satisfied.
# It is still recommended to specify flat=False if non-flat output is intend.
print(auto_un_flat()(1))  # will print the value 1, 2 or 3, mind the double-double parenthesis
```

#### Mixing Static Objects with Callable Objects
```python
from Fortuna import TruffleShuffle


""" With automatic flattening active, lambda() -> int can be treated as an int. """
mixed_flat = TruffleShuffle([1, 2, lambda: 3])
print(mixed_flat())  # will print 1, 2 or 3

mixed_un_flat = TruffleShuffle([1, 2, lambda: 3], flat=False) # this pattern is not recommended.
print(mixed_flat())  # will print 1, 2 or "Function <lambda at some_address>"
# This pattern is not recommended because you wont know the nature of what you get back.
# This is almost always not what you want, and it can give rise to messy logic in other areas of your code.
```

#### Dynamic Strings
To successfully express a dynamic string, and keep it dynamic, at least one level of indirection is required. Without an indirection the f-string would collapse into a static string too soon.

```python
from Fortuna import TruffleShuffle, d


# d() is a simple dice function, d(n) -> [1, n] flat uniform distribution.
dynamic_string = TruffleShuffle((
    # while the probability of all A == all B == all C, individual probabilities of each possible string will differ based on the number of possible outputs of each category.
    lambda: f"A{d(2)}",  # -> A1 - A2, each are twice as likely as any particular B, and three times as likely as any C.
    lambda: f"B{d(4)}",  # -> B1 - B4, each are half as likely as any particular A, and 3/2 as likely as any C.
    lambda: f"C{d(6)}",  # -> C1 - C6, each are 1/3 as likely as any particular A and 2/3 as likely of any B.
))

print(dynamic_string())  # prints a random dynamic string, flattened at call time.

"""
>>> distribution_timer(dynamic_string)
Output Analysis: TruffleShuffle(sequence_data)()
Typical Timing: 875 ± 8 ns
Distribution of 10000 Samples:
 A1: 16.9%
 A2: 16.43%
 B1: 8.39%
 B2: 8.21%
 B3: 8.28%
 B4: 8.46%
 C1: 5.25%
 C2: 6.09%
 C3: 5.59%
 C4: 5.32%
 C5: 5.41%
 C6: 5.67%
"""
```

#### Nesting Dolls
```python
from Fortuna import TruffleShuffle


nesting_dolls = TruffleShuffle({
    TruffleShuffle({"A", "B", "C", "D", "E"}),
    TruffleShuffle({"F", "G", "H", "I", "J"}),
    TruffleShuffle({"K", "L", "M", "N", "O"}),
    TruffleShuffle({"P", "Q", "R", "S", "T"}),
})

print(nesting_dolls())  # prints one of the letters A-T, double wide distribution.
```

### QuantumMonty
`Fortuna.QuantumMonty(data: Collection, flat=True) -> Callable -> Value`
- @param data :: Collection of Values.
- @param flat :: Bool. Default: True. Option to flatten callable values.
- @return :: Callable Object with Monty Methods for producing various distributions of the data.
    - @return :: Random value from the data. The instance will produce random values from the list using the selected distribution model or "monty". The default monty is the Quantum Monty Algorithm.

```python
from Fortuna import QuantumMonty


list_of_values = [1, 2, 3, 4, 5, 6]
monty = QuantumMonty(list_of_values)

print(monty())               # prints a random value from the list_of_values.
                             # uses the default Quantum Monty Algorithm.

print(monty.flat_uniform())  # prints a random value from the list_of_values.
                             # uses the "uniform" monty: a flat uniform distribution.
                             # equivalent to random.choice(list_of_values).
```
The QuantumMonty class represents a diverse collection of strategies for producing random values from a sequence where the output distribution is based on the method you choose. Generally speaking, each value in the sequence will have a probability that is based on its position in the sequence. For example: the "front" monty produces random values where the beginning of the sequence is geometrically more common than the back. Given enough samples the "front" monty will always converge to a 45 degree slope down for any list of unique values.

There are three primary method families: linear, gaussian, and poisson. Each family has three base methods; 'front', 'middle', 'back', plus a 'quantum' method that incorporates all three base methods. The quantum algorithms for each family produce distributions by overlapping the probability waves of the other methods in their family. The Quantum Monty Algorithm incorporates all nine base methods.

```python
import Fortuna


monty = Fortuna.QuantumMonty(
    ["Alpha", "Beta", "Delta", "Eta", "Gamma", "Kappa", "Zeta"]
)

# Each of the following methods will return a random value from the sequence.
# Each method has its own unique distribution model for the same data set.

""" Flat Base Case """
monty.flat_uniform()             # Flat Uniform Distribution

""" Geometric Positional """
monty.front_linear()        # Linear Descending, Triangle
monty.middle_linear()       # Linear Median Peak, Equilateral Triangle
monty.back_linear()         # Linear Ascending, Triangle
monty.quantum_linear()      # Linear Overlay, 3-way monty.

""" Gaussian Positional """
monty.front_gauss()         # Front Gamma
monty.middle_gauss()        # Scaled Gaussian
monty.back_gauss()          # Reversed Gamma
monty.quantum_gauss()       # Gaussian Overlay, 3-way monty.

""" Poisson Positional """
monty.front_poisson()       # 1/4 Mean Poisson
monty.middle_poisson()      # 1/2 Mean Poisson
monty.back_poisson()        # 3/4 Mean Poisson
monty.quantum_poisson()     # Poisson Overlay, 3-way monty.

""" Quantum Monty Algorithm """
monty()                     # Quantum Monty Algorithm, 9-way monty.
monty.quantum_monty()
```

### Weighted Choice: Base Class
Weighted Choice offers two strategies for selecting random values from a sequence where programmable rarity is desired. Both produce a custom distribution of values based on the weights of the values.

The choice to use one strategy over the other is purely about which one suits you or your data best. Relative weights are easier to understand at a glance. However, many RPG Treasure Tables map rather nicely to a cumulative weighted strategy.

#### Cumulative Weighted Choice
`Fortuna.CumulativeWeightedChoice(weighted_table: Table, flat=True) -> Callable -> Value`
- @param weighted_table :: Table of weighted pairs.
- @param flat :: Bool. Default: True. Option to flatten callable values.
- @return :: Callable Instance
    - @return :: Random value from the weighted_table, distribution based on the weights of the values.

_Note: Logic dictates Cumulative Weights must be unique!_

```python
from Fortuna import CumulativeWeightedChoice


cum_weighted_choice = CumulativeWeightedChoice([
    (7, "Apple"),
    (11, "Banana"),
    (13, "Cherry"),
    (23, "Grape"),
    (26, "Lime"),
    (30, "Orange"),  # same as relative weight 4 because 30 - 26 = 4
])

print(cum_weighted_choice())  # prints a weighted random value
```

#### Relative Weighted Choice
`Fortuna.RelativeWeightedChoice(weighted_table: Table) -> Callable -> Value`
- @param weighted_table :: Table of weighted pairs.
- @param flat :: Bool. Default: True. Option to flatten callable values.
- @return :: Callable Instance
    - @return :: Random value from the weighted_table, distribution based on the weights of the values.

```python
from Fortuna import RelativeWeightedChoice


population = ["Apple", "Banana", "Cherry", "Grape", "Lime", "Orange"]
rel_weights = [7, 4, 2, 10, 3, 4]
rel_weighted_choice = RelativeWeightedChoice(zip(rel_weights, population))

print(rel_weighted_choice())  # prints a weighted random value
```

### FlexCat
`Fortuna.FlexCat(dict_of_lists: Matrix, key_bias="front_linear", val_bias="truffle_shuffle", flat=True) -> Callable -> Value`
- @param dict_of_lists :: Keyed Matrix of Value Sequences.
- @parm key_bias :: String indicating the name of the algorithm to use for random key selection.
- @parm val_bias :: String indicating the name of the algorithm to use for random value selection.
- @param flat :: Bool. Default: True. Option to flatten callable values.
- @return :: Callable Instance
    - @param cat_key :: Optional String. Default is None. Key selection by name. If specified, this will override the key_bias for a single call.
    - @return :: Value. Returns a random value generated with val_bias from a random sequence generated with key_bias.

FlexCat is like a two dimensional QuantumMonty, or a QuantumMonty of QuantumMontys.

The constructor takes two optional keyword arguments to specify the algorithms to be used to make random selections. The algorithm specified for selecting a key need not be the same as the one for selecting values. An optional key may be provided at call time to bypass the random key selection. Keys passed in this way must exactly match a key in the Matrix.

By default, FlexCat will use key_bias="front_linear" and val_bias="truffle_shuffle", this will make the top of the data structure geometrically more common than the bottom and it will truffle shuffle the sequence values. This config is known as TopCat, it produces a descending-step, micro-shuffled distribution sequence. Many other combinations are available.

Algorithmic Options: _See QuantumMonty & TruffleShuffle for more details._
- "front_linear", Linear Descending
- "middle_linear", Linear Median Peak
- "back_linear", Linear Ascending
- "quantum_linear", Linear 3-way monty
- "front_gauss", Gamma Descending
- "middle_gauss", Scaled Gaussian
- "back_gauss", Gamma Ascending
- "quantum_gauss", Gaussian 3-way monty
- "front_poisson", Front 1/3 Mean Poisson
- "middle_poisson", Middle Mean Poisson
- "back_poisson", Back 1/3 Mean Poisson
- "quantum_poisson", Poisson 3-way monty
- "quantum_monty", Quantum Monty Algorithm, 9-way monty
- "flat_uniform", uniform flat distribution
- "truffle_shuffle", TruffleShuffle, wide uniform distribution

```python
from Fortuna import FlexCat


matrix_data = {
    "Cat_A": (f"A{i}" for i in range(1, 6)),
    "Cat_B": ("B1", "B2", "B3", "B4", "B5"),
    "Cat_C": ("C1", "C2", "C3", "C4", "C5"),
}
flex_cat = FlexCat(matrix_data, key_bias="front_linear", val_bias="flat_uniform")

flex_cat()  # returns a "flat_uniform" random value from a random "front_linear" weighted category.
flex_cat("Cat_B")  # returns a "flat_uniform" random value specifically from the "Cat_B" category of Values.
```

### Random Integer Generators
`Fortuna.random_below(number: int) -> int`
- @param number :: Any Integer
- @return :: Returns a random integer in the range...
    - `random_below(number) -> [0, number)` for positive values.
    - `random_below(number) -> (number, 0]` for negative values.
    - `random_below(0) -> 0` Always returns zero when input is zero
- Flat uniform distribution.


`Fortuna.random_int(left_limit: int, right_limit: int) -> int`
- @param left_limit :: Any Integer
- @param right_limit :: Any Integer
- @return :: Returns a random integer in the range [left_limit, right_limit]
    - `random_int(1, 10) -> [1, 10]`
    - `random_int(10, 1) -> [1, 10]` same as above.
    - `random_int(A, B)` Always returns A when A == B
- Flat uniform distribution.


`Fortuna.random_range(start: int, stop: int = 0, step: int = 1) -> int`
- @param start :: Required starting point.
    - `random_range(0) -> [0]`
    - `random_range(10) -> [0, 10)` from 0 to 9. Same as `Fortuna.random_index(N)`
    - `random_range(-10) -> [-10, 0)` from -10 to -1. Same as `Fortuna.random_index(-N)`
- @param stop :: Zero by default. Optional range bound. With at least two arguments, the order of the first two does not matter.
    - `random_range(0, 0) -> [0]`
    - `random_range(0, 10) -> [0, 10)` from 0 to 9.
    - `random_range(10, 0) -> [0, 10)` same as above.
- @param step :: One by default. Optional step size.
    - `random_range(0, 0, 0) -> [0]`
    - `random_range(0, 10, 2) -> [0, 10) by 2` even numbers from 0 to 8.
    - The sign of the step parameter controls the phase of the output. Negative stepping will flip the inclusively.
    - `random_range(0, 10, -1) -> (0, 10]` starts at 10 and ranges down to 1.
    - `random_range(10, 0, -1) -> (0, 10]` same as above.
    - `random_range(10, 10, 0) -> [10]` a step size or range size of zero always returns the first parameter.
- @return :: Returns a random integer in the range [A, B) by increments of C.
- Flat uniform distribution.


`Fortuna.d(sides: int) -> int`
- Represents a single roll of a given size die.
- @param sides :: Any Integer. Represents the size or number of sides, most commonly six.
- @return :: Returns a random integer in the range [1, sides].
- Flat uniform distribution.


`Fortuna.dice(rolls: int, sides: int) -> int`
- Represents the sum of multiple rolls of the same size die.
- @param rolls :: Any Integer. Represents the number of times to roll the die.
- @param sides :: Any Integer. Represents the die size or number of sides, most commonly six.
- @return :: Returns a random integer in range [X, Y] where X = rolls and Y = rolls * sides.
- Geometric distribution based on the number and size of the dice rolled.
- Complexity scales primarily with the number of rolls, not the size of the dice.


`Fortuna.plus_or_minus(number: int) -> int`
- @param number :: Any Integer.
- @return :: Returns a random integer in range [-number, number].
- Flat uniform distribution.


`Fortuna.plus_or_minus_linear(number: int) -> int`
- @param number :: Any Integer.
- @return :: Returns a random integer in range [-number, number].
- Linear geometric, 45 degree triangle distribution centered on zero.


`Fortuna.plus_or_minus_gauss(number: int) -> int`
- @param number :: Any Integer.
- @return :: Returns a random integer in range [-number, number].
- Stretched gaussian distribution centered on zero.


### Random Index, ZeroCool Specification
ZeroCool Methods are used to generate random Sequence indices.

ZeroCool methods must have the following properties:
- Any distribution model is acceptable.
- The method or function must take exactly one Integer parameter N.
- The method returns a random int in range `[0, N)` for positive values of N.
- The method returns a random int in range `[N, 0)` for negative values of N.
    - This symmetry matches how python can index a list from the back for negative index values or from the front for positive index values.

ZeroCool functions often have an interesting limit as size goes to zero. ZeroCool does not place requirements the output of this input limit. At higher levels of abstraction, inside classes that employ ZeroCool methods- zero is always a sentinel to indicate the full range. In that case the length of the list is sent to the ZeroCool method, not zero. However for those who enjoy thinking a little deeper, consider the following:

If given the fact that an empty range is never an option, we could design a better solution than failure for input zero. Calculus might suggest that both infinity and negative infinity are equally viable output for an input limit of zero, but both are inappropriate for indexing a list. However, if we map infinity to the back of the list and minus infinity to the front of the list, then the following might hold: `random_index(0) -> [-1, 0]`. This "Top or Bottom" solution is not required for a method to be ZeroCool compatible, it is just an interesting option. Other valid possibilities include: always return None or 0 or -1 or throw an exception or spawn nasal demons, however none of these seem terribly helpful or useful. At least the Top/Bottom solution accurately reflects the "off by one" symmetry of the domain mapping that defines ZeroCool methods in general.


```python
from Fortuna import random_index


some_list = [i for i in range(100)]

print(some_list[random_index(10)])  # prints one of the first 10 items of some_list, [0, 9]
print(some_list[random_index(-10)])  # prints one of the last 10 items of some_list, [90, 99]
```
### ZeroCool Methods
- `Fortuna.random_index(size: int) -> int` Flat uniform distribution
- `Fortuna.front_gauss(size: int) -> int` Gamma Distribution: Front Peak
- `Fortuna.middle_gauss(size: int) -> int` Stretched Gaussian Distribution: Median Peak
- `Fortuna.back_gauss(size: int) -> int` Gamma Distribution: Back Peak
- `Fortuna.quantum_gauss(size: int) -> int` Quantum Gaussian: Three-way Monty
- `Fortuna.front_poisson(size: int) -> int` Poisson Distribution: Front 1/3 Peak
- `Fortuna.middle_poisson(size: int) -> int` Poisson Distribution: Middle Peak
- `Fortuna.back_poisson(size: int) -> int` Poisson Distribution: Back 1/3 Peak
- `Fortuna.quantum_poisson(size: int) -> int` Quantum Poisson: Three-way Monty
- `Fortuna.front_geometric(size: int) -> int` Linear Geometric: 45 Degree Front Peak
- `Fortuna.middle_geometric(size: int) -> int` Linear Geometric: 45 Degree Middle Peak
- `Fortuna.back_geometric(size: int) -> int` Linear Geometric: 45 Degree Back Peak
- `Fortuna.quantum_geometric(size: int) -> int` Quantum Geometric: Three-way Monty
- `Fortuna.quantum_monty(size: int) -> int` Quantum Monty: Twelve-way Monty

```python
from Fortuna import front_gauss, middle_gauss, back_gauss, quantum_gauss


some_list = [i for i in range(100)]

# Each of the following prints one of the first 10 items of some_list with the appropriate distribution
print(some_list[front_gauss(10)])
print(some_list[middle_gauss(10)])
print(some_list[back_gauss(10)])
print(some_list[quantum_gauss(10)])

# Each of the following prints one of the last 10 items of some_list with the appropriate distribution
print(some_list[front_gauss(-10)])  
print(some_list[middle_gauss(-10)])  
print(some_list[back_gauss(-10)])  
print(some_list[quantum_gauss(-10)])
```

### Random Float Generator
`Fortuna.canonical() -> float`
- @return :: random float in range [0.0, 1.0), flat uniform.


`Fortuna.random_float(a: Float, b: Float) -> Float`
- @param a :: Float
- @param b :: Float
- @return :: random Float in range [a, b), flat uniform distribution.


### Random Truth Generator
`Fortuna.percent_true(truth_factor: Float = 50.0) -> bool`
- @param truth_factor :: The probability of True as a percentage. Default is 50 percent.
- @return :: Produces True or False based on the truth_factor.
    - Always returns False if num is 0 or less
    - Always returns True if num is 100 or more.


### Shuffle Algorithms
`Fortuna.shuffle(array: list) -> None`
- Knuth B shuffle algorithm. Destructive, in-place shuffle.
- @param array :: Must be a mutable list.

`Fortuna.knuth(array: list) -> None`
- Knuth A shuffle algorithm. Destructive, in-place shuffle.
- @param array :: Must be a mutable list.

`Fortuna.fisher_yates(array: list) -> None`
- Fisher-Yates shuffle algorithm. Destructive, in-place shuffle.
- @param array :: Must be a mutable list.


### Test Suite
`Fortuna.distribution_timer(func: staticmethod, *args, num_cycles=100000, **kwargs) -> None`

`Fortuna.quick_test(num_cycles=10000) -> None`


## Fortuna Development Log
##### Fortuna 3.6.2
- Installer Script Update

##### Fortuna 3.6.1
- Documentation Update

##### Fortuna 3.6.0
- Storm Update
- Test Update
- Bug fix for random_range(), negative stepping is now working as intended. This bug was introduced in 3.5.0.
- 25% performance gain: shuffle() function. Now 25 times faster than Random.shuffle().
- Removed Features
    - lazy_cat(): use QuantumMonty class instead.
    - flex_cat(): use FlexCat class instead.
    - truffle_shuffle(): use TruffleShuffle class instead.

##### Fortuna 3.5.3 - internal
- Features added for testing & development
    - ActiveChoice class
    - random_rotate() function

##### Fortuna 3.5.2
- Documentation Updates

##### Fortuna 3.5.1
- Test Update

##### Fortuna 3.5.0
- Storm Update
- Minor Bug Fix: Truffle Shuffle

##### Fortuna 3.4.9
- Test Update

##### Fortuna 3.4.8
- Storm Update

##### Fortuna 3.4.7
- Bug fix for analytic_continuation.

##### Fortuna 3.4.6
- Docs Update

##### Fortuna 3.4.5
- Docs Update
- Range Tests Added, see extras folder.

##### Fortuna 3.4.4
- ZeroCool Algorithm Bug Fixes
- Typos Fixed

##### Fortuna 3.4.3
- Docs Update

##### Fortuna 3.4.2
- Typos Fixed

##### Fortuna 3.4.1
- Major Bug Fix: random_index()

##### Fortuna 3.4.0 - internal
- ZeroCool Poisson Algorithm Family Updated

##### Fortuna 3.3.8 - internal
- Docs Update

##### Fortuna 3.3.7
- Fixed Performance Bug: ZeroCool Linear Algorithm Family

##### Fortuna 3.3.6
- Docs Update

##### Fortuna 3.3.5
- ABI Updates
- Bug Fixes

##### Fortuna 3.3.4
- Examples Update

##### Fortuna 3.3.3
- Test Suite Update

##### Fortuna 3.3.2 - internal
- Documentation Update

##### Fortuna 3.3.1 - internal
- Minor Bug Fix

##### Fortuna 3.3.0 - internal
- Added `plus_or_minus_gauss(N: int) -> int` random int in range [-N, N] Stretched Gaussian Distribution

##### Fortuna 3.2.3
- Small Typos Fixed

##### Fortuna 3.2.2
- Documentation update.

##### Fortuna 3.2.1
- Small Typo Fixed

##### Fortuna 3.2.0
- API updates:
    - QunatumMonty.uniform -> QunatumMonty.flat_uniform
    - QunatumMonty.front -> QunatumMonty.front_linear
    - QunatumMonty.middle -> QunatumMonty.middle_linear
    - QunatumMonty.back -> QunatumMonty.back_linear
    - QunatumMonty.quantum -> QunatumMonty.quantum_linear
    - randindex -> random_index
    - randbelow -> random_below
    - randrange -> random_range
    - randint   -> random_int

##### Fortuna 3.1.0
- `discrete()` has been removed, see Weighted Choice.
- `lazy_cat()` added.
- All ZeroCool methods have been raised to top level API, for use with lazy_cat()

##### Fortuna 3.0.1
- minor typos.

##### Fortuna 3.0.0
- Storm 2 Rebuild.

##### Fortuna 2.1.1
- Small bug fixes.
- Test updates.

##### Fortuna 2.1.0, Major Feature Update
- Fortuna now includes the best of RNG and Pyewacket.

##### Fortuna 2.0.3
- Bug fix.

##### Fortuna 2.0.2
- Clarified some documentation.

##### Fortuna 2.0.1
- Fixed some typos.

##### Fortuna 2.0.0b1-10
- Total rebuild. New RNG Storm Engine.

##### Fortuna 1.26.7.1
- README updated.

##### Fortuna 1.26.7
- Small bug fix.

##### Fortuna 1.26.6
- Updated README to reflect recent changes to the test script.

##### Fortuna 1.26.5
- Fixed small bug in test script.

##### Fortuna 1.26.4
- Updated documentation for clarity.
- Fixed a minor typo in the test script.

##### Fortuna 1.26.3
- Clean build.

##### Fortuna 1.26.2
- Fixed some minor typos.

##### Fortuna 1.26.1
- Release.

##### Fortuna 1.26.0 beta 2
- Moved README and LICENSE files into fortuna_extras folder.

##### Fortuna 1.26.0 beta 1
- Dynamic version scheme implemented.
- The Fortuna Extension now requires the fortuna_extras package, previously it was optional.

##### Fortuna 1.25.4
- Fixed some minor typos in the test script.

##### Fortuna 1.25.3
- Since version 1.24 Fortuna requires Python 3.7 or higher. This patch corrects an issue where the setup script incorrectly reported requiring Python 3.6 or higher.

##### Fortuna 1.25.2
- Updated test suite.
- Major performance update for TruffleShuffle.
- Minor performance update for QuantumMonty & FlexCat: cycle monty.

##### Fortuna 1.25.1
- Important bug fix for TruffleShuffle, QuantumMonty and FlexCat.

##### Fortuna 1.25
- Full 64bit support.
- The Distribution & Performance Tests have been redesigned.
- Bloat Control: Two experimental features have been removed.
    - RandomWalk
    - CatWalk
- Bloat Control: Several utility functions have been removed from the top level API. These function remain in the Fortuna namespace for now, but may change in the future without warning.
    - stretch_bell, internal only.
    - min_max, not used anymore.
    - analytic_continuation, internal only.
    - flatten, internal only.

##### Fortuna 1.24.3
- Low level refactoring, non-breaking patch.

##### Fortuna 1.24.2
- Setup config updated to improve installation.

##### Fortuna 1.24.1
- Low level patch to avoid potential ADL issue. All low level function calls are now qualified.

##### Fortuna 1.24
- Documentation updated for even more clarity.
- Bloat Control: Two naïve utility functions that are no longer used in the module have been removed.
    - n_samples -> use a list comprehension instead. `[f(x) for _ in range(n)]`
    - bind -> use a lambda instead. `lambda: f(x)`

##### Fortuna 1.23.7
- Documentation updated for clarity.
- Minor bug fixes.
- TruffleShuffle has been redesigned slightly, it now uses a random rotate instead of swap.
- Custom `__repr__` methods have been added to each class.

##### Fortuna 1.23.6
- New method for QuantumMonty: quantum_not_monty - produces the upside down quantum_monty.
- New bias option for FlexCat: not_monty.

##### Fortuna 1.23.5.1
- Fixed some small typos.

##### Fortuna 1.23.5
- Documentation updated for clarity.
- All sequence wrappers can now accept generators as input.
- Six new functions added:
    - random_float() -> float in range [0.0..1.0) exclusive, uniform flat distribution.
    - percent_true_float(num: float) -> bool, Like percent_true but with floating point precision.
    - plus_or_minus_linear_down(num: int) -> int in range [-num..num], upside down pyramid.
    - plus_or_minus_curve_down(num: int) -> int in range [-num..num], upside down bell curve.
    - mostly_not_middle(num: int) -> int in range [0..num], upside down pyramid.
    - mostly_not_center(num: int) -> int in range [0..num], upside down bell curve.
- Two new methods for QuantumMonty:
    - mostly_not_middle
    - mostly_not_center
- Two new bias options for FlexCat, either can be used to define x and/or y axis bias:
    - not_middle
    - not_center

##### Fortuna 1.23.4.2
- Fixed some minor typos in the README.md file.

##### Fortuna 1.23.4.1
- Fixed some minor typos in the test suite.

##### Fortuna 1.23.4
- Fortuna is now Production/Stable!
- Fortuna and Fortuna Pure now use the same test suite.

##### Fortuna 0.23.4, first release candidate.
- RandomCycle, BlockCycle and TruffleShuffle have been refactored and combined into one class: TruffleShuffle.
- QuantumMonty and FlexCat will now use the new TruffleShuffle for cycling.
- Minor refactoring across the module.

##### Fortuna 0.23.3, internal
- Function shuffle(arr: list) added.

##### Fortuna 0.23.2, internal
- Simplified the plus_or_minus_curve(num: int) function, output will now always be bounded to the range [-num..num].
- Function stretched_bell(num: int) added, this matches the previous behavior of an unbounded plus_or_minus_curve.

##### Fortuna 0.23.1, internal
- Small bug fixes and general clean up.

##### Fortuna 0.23.0
- The number of test cycles in the test suite has been reduced to 10,000 (down from 100,000). The performance of the pure python implementation and the c-extension are now directly comparable.
- Minor tweaks made to the examples in `.../fortuna_extras/fortuna_examples.py`

##### Fortuna 0.22.2, experimental features
- BlockCycle class added.
- RandomWalk class added.
- CatWalk class added.

##### Fortuna 0.22.1
- Fortuna classes no longer return lists of values, this behavior has been extracted to a free function called n_samples.

##### Fortuna 0.22.0, experimental features
- Function bind added.
- Function n_samples added.

##### Fortuna 0.21.3
- Flatten will no longer raise an error if passed a callable item that it can't call. It correctly returns such items in an uncalled state without error.
- Simplified `.../fortuna_extras/fortuna_examples.py` - removed unnecessary class structure.

##### Fortuna 0.21.2
- Fix some minor bugs.

##### Fortuna 0.21.1
- Fixed a bug in `.../fortuna_extras/fortuna_examples.py`

##### Fortuna 0.21.0
- Function flatten added.
- Flatten: The Fortuna classes will recursively unpack callable objects in the data set.

##### Fortuna 0.20.10
- Documentation updated.

##### Fortuna 0.20.9
- Minor bug fixes.

##### Fortuna 0.20.8, internal
- Testing cycle for potential new features.

##### Fortuna 0.20.7
- Documentation updated for clarity.

##### Fortuna 0.20.6
- Tests updated based on recent changes.

##### Fortuna 0.20.5, internal
- Documentation updated based on recent changes.

##### Fortuna 0.20.4, internal
- WeightedChoice (both types) can optionally return a list of samples rather than just one value, control the length of the list via the n_samples argument.

##### Fortuna 0.20.3, internal
- RandomCycle can optionally return a list of samples rather than just one value,
control the length of the list via the n_samples argument.

##### Fortuna 0.20.2, internal
- QuantumMonty can optionally return a list of samples rather than just one value,
control the length of the list via the n_samples argument.

##### Fortuna 0.20.1, internal
- FlexCat can optionally return a list of samples rather than just one value,
control the length of the list via the n_samples argument.

##### Fortuna 0.20.0, internal
- FlexCat now accepts a standard dict as input. The ordered(ness) of dict is now part of the standard in Python 3.7.1. Previously FlexCat required an OrderedDict, now it accepts either and treats them the same.

##### Fortuna 0.19.7
- Fixed bug in `.../fortuna_extras/fortuna_examples.py`.

##### Fortuna 0.19.6
- Updated documentation formatting.
- Small performance tweak for QuantumMonty and FlexCat.

##### Fortuna 0.19.5
- Minor documentation update.

##### Fortuna 0.19.4
- Minor update to all classes for better debugging.

##### Fortuna 0.19.3
- Updated plus_or_minus_curve to allow unbounded output.

##### Fortuna 0.19.2
- Internal development cycle.
- Minor update to FlexCat for better debugging.

##### Fortuna 0.19.1
- Internal development cycle.

##### Fortuna 0.19.0
- Updated documentation for clarity.
- MultiCat has been removed, it is replaced by FlexCat.
- Mostly has been removed, it is replaced by QuantumMonty.

##### Fortuna 0.18.7
- Fixed some more README typos.

##### Fortuna 0.18.6
- Fixed some README typos.

##### Fortuna 0.18.5
- Updated documentation.
- Fixed another minor test bug.

##### Fortuna 0.18.4
- Updated documentation to reflect recent changes.
- Fixed some small test bugs.
- Reduced default number of test cycles to 10,000 - down from 100,000.

##### Fortuna 0.18.3
- Fixed some minor README typos.

##### Fortuna 0.18.2
- Fixed a bug with Fortuna Pure.

##### Fortuna 0.18.1
- Fixed some minor typos.
- Added tests for `.../fortuna_extras/fortuna_pure.py`

##### Fortuna 0.18.0
- Introduced new test format, now includes average call time in nanoseconds.
- Reduced default number of test cycles to 100,000 - down from 1,000,000.
- Added pure Python implementation of Fortuna: `.../fortuna_extras/fortuna_pure.py`
- Promoted several low level functions to top level.
    - `zero_flat(num: int) -> int`
    - `zero_cool(num: int) -> int`
    - `zero_extreme(num: int) -> int`
    - `max_cool(num: int) -> int`
    - `max_extreme(num: int) -> int`
    - `analytic_continuation(func: staticmethod, num: int) -> int`
    - `min_max(num: int, lo: int, hi: int) -> int`

##### Fortuna 0.17.3
- Internal development cycle.

##### Fortuna 0.17.2
- User Requested: dice() and d() functions now support negative numbers as input.

##### Fortuna 0.17.1
- Fixed some minor typos.

##### Fortuna 0.17.0
- Added QuantumMonty to replace Mostly, same default behavior with more options.
- Mostly is depreciated and may be removed in a future release.
- Added FlexCat to replace MultiCat, same default behavior with more options.
- MultiCat is depreciated and may be removed in a future release.
- Expanded the Treasure Table example in `.../fortuna_extras/fortuna_examples.py`

##### Fortuna 0.16.2
- Minor refactoring for WeightedChoice.

##### Fortuna 0.16.1
- Redesigned fortuna_examples.py to feature a dynamic random magic item generator.
- Raised cumulative_weighted_choice function to top level.
- Added test for cumulative_weighted_choice as free function.
- Updated MultiCat documentation for clarity.

##### Fortuna 0.16.0
- Pushed distribution_timer to the .pyx layer.
- Changed default number of iterations of tests to 1 million, up form 1 hundred thousand.
- Reordered tests to better match documentation.
- Added Base Case Fortuna.fast_rand_below.
- Added Base Case Fortuna.fast_d.
- Added Base Case Fortuna.fast_dice.

##### Fortuna 0.15.10
- Internal Development Cycle

##### Fortuna 0.15.9
- Added Base Cases for random_value()
- Added Base Case for randint()

##### Fortuna 0.15.8
- Clarified MultiCat Test

##### Fortuna 0.15.7
- Fixed minor typos.

##### Fortuna 0.15.6
- Fixed minor typos.
- Simplified MultiCat example.

##### Fortuna 0.15.5
- Added MultiCat test.
- Fixed some minor typos in docs.

##### Fortuna 0.15.4
- Performance optimization for both WeightedChoice() variants.
- Cython update provides small performance enhancement across the board.
- Compilation now leverages Python3 all the way down.
- MultiCat pushed to the .pyx layer for better performance.

##### Fortuna 0.15.3
- Reworked the MultiCat example to include several randomizing strategies working in concert.
- Added Multi Dice 10d10 performance tests.
- Updated sudo code in documentation to be more pythonic.

##### Fortuna 0.15.2
- Fixed: Linux installation failure.
- Added: complete source files to the distribution (.cpp .hpp .pyx).

##### Fortuna 0.15.1
- Updated & simplified distribution_timer in `fortuna_tests.py`
- Readme updated, fixed some typos.
- Known issue preventing successful installation on some linux platforms.

##### Fortuna 0.15.0
- Performance tweaks.
- Readme updated, added some details.

##### Fortuna 0.14.1
- Readme updated, fixed some typos.

##### Fortuna 0.14.0
- Fixed a bug where the analytic continuation algorithm caused a rare issue during compilation on some platforms.

##### Fortuna 0.13.3
- Fixed Test Bug: percent sign was missing in output distributions.
- Readme updated: added update history, fixed some typos.

##### Fortuna 0.13.2
- Readme updated for even more clarity.

##### Fortuna 0.13.1
- Readme updated for clarity.

##### Fortuna 0.13.0
- Minor Bug Fixes.
- Readme updated for aesthetics.
- Added Tests: `.../fortuna_extras/fortuna_tests.py`

##### Fortuna 0.12.0
- Internal test for future update.

##### Fortuna 0.11.0
- Initial Release: Public Beta

##### Fortuna 0.10.0
- Module name changed from Dice to Fortuna

##### Dice 0.1.x - 0.9.x
- Experimental Phase


## Fortuna Distribution and Performance Test Suite
```
Fortuna Test Suite: RNG Storm Engine

Output Range Tests: Success!

=========================================================================

Fortuna Quick Test

Random Sequence Values:

some_list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

Base Case
Output Analysis: Random.choice(some_list)
Typical Timing: 719 ± 10 ns
Statistics of 1000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.412
 Std Deviation: 2.8762003630707875
Distribution of 10000 Samples:
 0: 10.42%
 1: 10.05%
 2: 9.6%
 3: 9.99%
 4: 10.3%
 5: 9.97%
 6: 9.62%
 7: 10.8%
 8: 9.65%
 9: 9.6%

Output Analysis: random_value(some_list)
Typical Timing: 157 ± 1 ns
Statistics of 1000 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.58
 Std Deviation: 2.862830032641791
Distribution of 10000 Samples:
 0: 9.75%
 1: 10.11%
 2: 10.12%
 3: 9.93%
 4: 9.91%
 5: 10.04%
 6: 10.11%
 7: 10.19%
 8: 9.89%
 9: 9.95%

Output Analysis: TruffleShuffle(sequence_data)()
Typical Timing: 375 ± 4 ns
Statistics of 1000 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.524
 Std Deviation: 2.8767515881110937
Distribution of 10000 Samples:
 0: 10.34%
 1: 9.97%
 2: 10.28%
 3: 10.0%
 4: 9.99%
 5: 10.02%
 6: 9.6%
 7: 9.83%
 8: 10.06%
 9: 9.91%

Output Analysis: QuantumMonty(sequence_data)()
Typical Timing: 313 ± 7 ns
Statistics of 1000 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.597
 Std Deviation: 2.953525742957781
Distribution of 10000 Samples:
 0: 10.56%
 1: 9.77%
 2: 8.73%
 3: 9.74%
 4: 11.03%
 5: 12.05%
 6: 9.11%
 7: 8.83%
 8: 9.13%
 9: 11.05%


Weighted Tables:

population = ('A', 'B', 'C', 'D')
cum_weights = (1, 3, 6, 10)  # partial_sum of rel_weights:  λ.xy: x + y
rel_weights = (1, 2, 3, 4)  # adjacent_difference of cum_weights:  λ.xy: x - y
cum_weighted_table = ((1, 'A'), (3, 'B'), (6, 'C'), (10, 'D'))  # or zip(cum_weights, population)
rel_weighted_table = ((1, 'A'), (2, 'B'), (3, 'C'), (4, 'D'))  # or zip(rel_weights, population)

Cumulative Base Case
Output Analysis: Random.choices(population, cum_weights=cum_weights)
Typical Timing: 1625 ± 8 ns
Distribution of 10000 Samples:
 A: 9.59%
 B: 19.49%
 C: 30.79%
 D: 40.13%

Output Analysis: CumulativeWeightedChoice(weighted_table)()
Typical Timing: 250 ± 8 ns
Distribution of 10000 Samples:
 A: 10.3%
 B: 19.93%
 C: 30.67%
 D: 39.1%

Relative Base Case
Output Analysis: Random.choices(population, weights=rel_weights)
Typical Timing: 2094 ± 8 ns
Distribution of 10000 Samples:
 A: 9.97%
 B: 20.2%
 C: 29.9%
 D: 39.93%

Output Analysis: RelativeWeightedChoice(weighted_table)()
Typical Timing: 250 ± 7 ns
Distribution of 10000 Samples:
 A: 10.0%
 B: 19.9%
 C: 30.54%
 D: 39.56%


Random Matrix Values:

some_matrix = {'A': (1, 2, 3, 4), 'B': (10, 20, 30, 40), 'C': (100, 200, 300, 400)}

Output Analysis: FlexCat(matrix_data, key_bias='flat_uniform', val_bias='flat_uniform')()
Typical Timing: 407 ± 5 ns
Statistics of 1000 Samples:
 Minimum: 1
 Median: 30
 Maximum: 400
 Mean: 95.862
 Std Deviation: 131.16781599139307
Distribution of 10000 Samples:
 1: 8.68%
 2: 8.23%
 3: 8.69%
 4: 8.48%
 10: 8.05%
 20: 8.19%
 30: 7.94%
 40: 8.21%
 100: 8.14%
 200: 8.5%
 300: 8.57%
 400: 8.32%


Random Integers:

Base Case
Output Analysis: Random.randrange(10)
Typical Timing: 813 ± 14 ns
Statistics of 1000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.401
 Std Deviation: 2.8838369432954276
Distribution of 10000 Samples:
 0: 10.18%
 1: 10.08%
 2: 10.14%
 3: 10.08%
 4: 9.91%
 5: 10.65%
 6: 9.79%
 7: 10.1%
 8: 9.59%
 9: 9.48%

Output Analysis: random_below(10)
Typical Timing: 63 ± 3 ns
Statistics of 1000 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.529
 Std Deviation: 2.8829621000752805
Distribution of 10000 Samples:
 0: 9.62%
 1: 10.61%
 2: 10.1%
 3: 9.92%
 4: 10.19%
 5: 10.45%
 6: 9.78%
 7: 9.69%
 8: 9.77%
 9: 9.87%

Output Analysis: random_index(10)
Typical Timing: 63 ± 4 ns
Statistics of 1000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.484
 Std Deviation: 2.8336855460290176
Distribution of 10000 Samples:
 0: 10.23%
 1: 10.3%
 2: 9.98%
 3: 9.76%
 4: 9.65%
 5: 10.1%
 6: 10.42%
 7: 9.73%
 8: 9.56%
 9: 10.27%

Output Analysis: random_range(10)
Typical Timing: 63 ± 8 ns
Statistics of 1000 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.317
 Std Deviation: 2.810410896899137
Distribution of 10000 Samples:
 0: 10.25%
 1: 10.0%
 2: 9.46%
 3: 9.64%
 4: 9.8%
 5: 10.13%
 6: 10.32%
 7: 10.11%
 8: 10.29%
 9: 10.0%

Output Analysis: random_below(-10)
Typical Timing: 63 ± 8 ns
Statistics of 1000 Samples:
 Minimum: -9
 Median: -4
 Maximum: 0
 Mean: -4.351
 Std Deviation: 2.8010077908932427
Distribution of 10000 Samples:
 -9: 9.7%
 -8: 9.69%
 -7: 9.48%
 -6: 10.22%
 -5: 10.63%
 -4: 9.91%
 -3: 10.28%
 -2: 9.92%
 -1: 10.07%
 0: 10.1%

Output Analysis: random_index(-10)
Typical Timing: 63 ± 8 ns
Statistics of 1000 Samples:
 Minimum: -10
 Median: -6
 Maximum: -1
 Mean: -5.625
 Std Deviation: 2.877960676878275
Distribution of 10000 Samples:
 -10: 9.99%
 -9: 10.11%
 -8: 10.01%
 -7: 9.8%
 -6: 10.13%
 -5: 9.74%
 -4: 9.82%
 -3: 10.27%
 -2: 10.22%
 -1: 9.91%

Output Analysis: random_range(-10)
Typical Timing: 94 ± 1 ns
Statistics of 1000 Samples:
 Minimum: -10
 Median: -5
 Maximum: -1
 Mean: -5.455
 Std Deviation: 2.927207465066927
Distribution of 10000 Samples:
 -10: 10.52%
 -9: 10.26%
 -8: 10.02%
 -7: 9.68%
 -6: 10.0%
 -5: 9.28%
 -4: 10.21%
 -3: 9.58%
 -2: 10.1%
 -1: 10.35%

Base Case
Output Analysis: Random.randrange(1, 10)
Typical Timing: 1032 ± 15 ns
Statistics of 1000 Samples:
 Minimum: 1
 Median: 5
 Maximum: 9
 Mean: 5.144
 Std Deviation: 2.59769359471282
Distribution of 10000 Samples:
 1: 10.92%
 2: 11.01%
 3: 10.87%
 4: 11.4%
 5: 11.61%
 6: 11.15%
 7: 10.5%
 8: 11.7%
 9: 10.84%

Output Analysis: random_range(1, 10)
Typical Timing: 63 ± 7 ns
Statistics of 1000 Samples:
 Minimum: 1
 Median: 5
 Maximum: 9
 Mean: 4.919
 Std Deviation: 2.636548954673588
Distribution of 10000 Samples:
 1: 11.19%
 2: 11.51%
 3: 10.9%
 4: 11.15%
 5: 11.23%
 6: 10.47%
 7: 11.48%
 8: 10.79%
 9: 11.28%

Output Analysis: random_range(10, 1)
Typical Timing: 63 ± 7 ns
Statistics of 1000 Samples:
 Minimum: 1
 Median: 5
 Maximum: 9
 Mean: 5.076
 Std Deviation: 2.647117532568441
Distribution of 10000 Samples:
 1: 10.92%
 2: 10.88%
 3: 11.82%
 4: 11.11%
 5: 11.26%
 6: 11.29%
 7: 10.67%
 8: 10.83%
 9: 11.22%

Base Case
Output Analysis: Random.randint(-5, 5)
Typical Timing: 1157 ± 11 ns
Statistics of 1000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: -0.183
 Std Deviation: 3.098889174222614
Distribution of 10000 Samples:
 -5: 8.93%
 -4: 8.96%
 -3: 9.05%
 -2: 9.02%
 -1: 9.27%
 0: 9.21%
 1: 9.65%
 2: 8.83%
 3: 9.26%
 4: 8.94%
 5: 8.88%

Output Analysis: random_int(-5, 5)
Typical Timing: 63 ± 5 ns
Statistics of 1000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: -0.045
 Std Deviation: 3.128380053919562
Distribution of 10000 Samples:
 -5: 8.82%
 -4: 9.03%
 -3: 9.12%
 -2: 9.47%
 -1: 8.75%
 0: 9.02%
 1: 10.04%
 2: 8.95%
 3: 8.76%
 4: 8.95%
 5: 9.09%

Base Case
Output Analysis: Random.randrange(1, 20, 2)
Typical Timing: 1282 ± 13 ns
Statistics of 1000 Samples:
 Minimum: 1
 Median: 9
 Maximum: 19
 Mean: 10.112
 Std Deviation: 5.70017087006591
Distribution of 10000 Samples:
 1: 10.09%
 3: 9.85%
 5: 10.24%
 7: 9.88%
 9: 9.96%
 11: 10.13%
 13: 9.28%
 15: 10.15%
 17: 10.0%
 19: 10.42%

Output Analysis: random_range(1, 20, 2)
Typical Timing: 63 ± 8 ns
Statistics of 1000 Samples:
 Minimum: 1
 Median: 9
 Maximum: 19
 Mean: 9.798
 Std Deviation: 5.760588544617675
Distribution of 10000 Samples:
 1: 10.03%
 3: 10.0%
 5: 10.18%
 7: 10.16%
 9: 9.8%
 11: 9.92%
 13: 10.56%
 15: 9.77%
 17: 9.93%
 19: 9.65%

Output Analysis: random_range(1, 20, -2)
Typical Timing: 94 ± 1 ns
Statistics of 1000 Samples:
 Minimum: 2
 Median: 10
 Maximum: 20
 Mean: 10.79
 Std Deviation: 5.779039478779955
Distribution of 10000 Samples:
 2: 10.21%
 4: 10.29%
 6: 9.98%
 8: 9.52%
 10: 10.69%
 12: 10.0%
 14: 9.6%
 16: 9.79%
 18: 9.55%
 20: 10.37%

Output Analysis: d(10)
Typical Timing: 63 ± 3 ns
Statistics of 1000 Samples:
 Minimum: 1
 Median: 5
 Maximum: 10
 Mean: 5.496
 Std Deviation: 2.82842429348639
Distribution of 10000 Samples:
 1: 9.51%
 2: 9.64%
 3: 10.12%
 4: 10.34%
 5: 10.63%
 6: 10.22%
 7: 9.51%
 8: 9.72%
 9: 9.81%
 10: 10.5%

Output Analysis: dice(3, 6)
Typical Timing: 94 ± 8 ns
Statistics of 1000 Samples:
 Minimum: 3
 Median: 10
 Maximum: 18
 Mean: 10.351
 Std Deviation: 2.93775925314336
Distribution of 10000 Samples:
 3: 0.41%
 4: 1.57%
 5: 2.62%
 6: 4.6%
 7: 6.57%
 8: 9.78%
 9: 11.9%
 10: 12.66%
 11: 11.89%
 12: 11.73%
 13: 10.08%
 14: 7.21%
 15: 4.28%
 16: 2.86%
 17: 1.41%
 18: 0.43%

Output Analysis: ability_dice(4)
Typical Timing: 188 ± 8 ns
Statistics of 1000 Samples:
 Minimum: 3
 Median: 13
 Maximum: 18
 Mean: 12.33
 Std Deviation: 2.8697273973907933
Distribution of 10000 Samples:
 3: 0.08%
 4: 0.32%
 5: 0.7%
 6: 1.55%
 7: 2.89%
 8: 4.84%
 9: 7.02%
 10: 9.09%
 11: 11.48%
 12: 12.63%
 13: 13.7%
 14: 12.65%
 15: 10.19%
 16: 7.08%
 17: 4.0%
 18: 1.78%

Output Analysis: plus_or_minus(5)
Typical Timing: 32 ± 3 ns
Statistics of 1000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: -0.117
 Std Deviation: 3.093684192331512
Distribution of 10000 Samples:
 -5: 9.13%
 -4: 9.08%
 -3: 9.5%
 -2: 9.59%
 -1: 8.85%
 0: 9.05%
 1: 8.9%
 2: 9.23%
 3: 8.97%
 4: 8.83%
 5: 8.87%

Output Analysis: plus_or_minus_linear(5)
Typical Timing: 63 ± 8 ns
Statistics of 1000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: -0.006
 Std Deviation: 2.3608340766639104
Distribution of 10000 Samples:
 -5: 2.72%
 -4: 5.62%
 -3: 7.99%
 -2: 12.01%
 -1: 14.06%
 0: 16.84%
 1: 13.2%
 2: 10.78%
 3: 8.08%
 4: 6.0%
 5: 2.7%

Output Analysis: plus_or_minus_gauss(5)
Typical Timing: 94 ± 8 ns
Statistics of 1000 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: 0.008
 Std Deviation: 1.6051518658720465
Distribution of 10000 Samples:
 -5: 0.14%
 -4: 1.2%
 -3: 4.47%
 -2: 11.86%
 -1: 20.23%
 0: 23.67%
 1: 20.93%
 2: 11.67%
 3: 4.61%
 4: 1.09%
 5: 0.13%


Random Floats:

Base Case
Output Analysis: Random.random()
Typical Timing: 32 ± 8 ns
Statistics of 1000 Samples:
 Minimum: 0.0009530611731990435
 Median: (0.5061621259960537, 0.5066632637028027)
 Maximum: 0.9999494629857831
 Mean: 0.49876081167796327
 Std Deviation: 0.29276480277392375
Post-processor Distribution of 10000 Samples using round method:
 0: 50.11%
 1: 49.89%

Output Analysis: canonical()
Typical Timing: 32 ± 8 ns
Statistics of 1000 Samples:
 Minimum: 0.000842323675811378
 Median: (0.5195936269238659, 0.5200284996845898)
 Maximum: 0.9972021814985803
 Mean: 0.5102006888999805
 Std Deviation: 0.28557283706466996
Post-processor Distribution of 10000 Samples using round method:
 0: 49.87%
 1: 50.13%

Output Analysis: random_float(0.0, 10.0)
Typical Timing: 32 ± 8 ns
Statistics of 1000 Samples:
 Minimum: 0.009450623552137476
 Median: (4.8487069636735525, 4.8503600028534)
 Maximum: 9.970645044844767
 Mean: 4.821798023219207
 Std Deviation: 2.8302244909343526
Post-processor Distribution of 10000 Samples using floor method:
 0: 10.17%
 1: 9.87%
 2: 9.67%
 3: 10.95%
 4: 10.21%
 5: 9.92%
 6: 9.62%
 7: 10.11%
 8: 10.09%
 9: 9.39%


Random Booleans:

Output Analysis: percent_true(33.33)
Typical Timing: 32 ± 8 ns
Statistics of 1000 Samples:
 Minimum: False
 Median: False
 Maximum: True
 Mean: 0.337
 Std Deviation: 0.4729214043101186
Distribution of 10000 Samples:
 False: 66.77%
 True: 33.23%


Shuffle Performance Tests:

some_med_list = [i for i in range(1000)]

Base Case: Random.shuffle(some_med_list)
Typical Timing: 706625 ± 1423 ns

fisher_yates(some_med_list)
Typical Timing: 72375 ± 209 ns

knuth(some_med_list)
Typical Timing: 72875 ± 45 ns

shuffle(some_med_list)  # default shuffle is the knuth_b algorithm
Typical Timing: 28375 ± 67 ns

-------------------------------------------------------------------------
Total Test Time: 1.376 seconds

```


## Legal Information
Fortuna © 2019 Broken aka Robert W Sharp, all rights reserved.

Fortuna is licensed under a Creative Commons Attribution-NonCommercial 3.0 Unported License.

See online version of this license here: <http://creativecommons.org/licenses/by-nc/3.0/>
